private import std.ctype;
private import std.stdio;
private import std.math;

/**

void string_to (
    {char,wchar,dchar} [] string,
    out {int,uint,long,ulong,float,double,real} value,
    type_flags flags = default_flags,
    out size_t end_offset = null);
        
These functions convert a string into an integer or floating-point
type, handling every form for C/C++, D, and C#, as well as some
extras.  You can also configure it to only handle a subset of the
functionality through the flags parameter.

These are the parameters:

    string is the string to convert; it can be a UTF-8, UTF-16, or
    UTF-32 string.

    value is the value to output to; it can be of type int, uint,
    long, ulong, float, double, or real.

    flags, if provided, sets the subset of numbers that can be
    converted.  This will be detailed later.  To act like various
    languages, you can use c_flags_int, c_flags_float, d_flags_int,
    d_flags_float, csharp_flags_int, or csharp_flags_float.

    end_offset, if provided, is stored with the array index that
    processing terminated on.

This can throw a set of errors:

    class_error_overflow is thrown if an overflow is detected when
    reading numbers.
    
    class_error_no_number_found is thrown if processing terminates
    without any digits read.
    
Here are the string formats which can be parsed:

    "    45" (unless if flags.whitespace is false)
    "-4", "+8" (unless if flags.sign is false)
    "0x8F", "0X5" (unless if flags.hex is false)
    "8Fh", "AH" (unless if flags.hex_suffix is false)
    "0770" (unless if flags.oct is false)
    "0B1011", "0b1111" (unless if flags.bin is false)
    "1_004" (unless if flags.underlines is false)
    "45e4" (float only - unless if flags.exp is false)
    "16^-8" (float only - unless if flags.exp_hexbin is false)
    
flags.exp_hexbin is crazy stuff for D.  Maybe it's normal to work
that way, I don't know.  flags.radix can be used to set the default radix if
none others are provided.

*/

class class_error_overflow : Error
{
    this ()
    {
        super ("Overflow has occurred while reading number.");
    }
}

class class_error_no_number_found : Error
{
    this ()
    {
        super ("No number was found.");
    }
}

/** The flags that can be passed to the string_to function. */

struct type_flags
{
    bit whitespace = true; /**< Whether to ignore whitespace at the beginning and end. */
    bit sign = true; /**< Whether to parse sign prefix (+ and -). */
    bit hex = true; /**< Whether to parse hexadecimal prefix (0x, 0X). */
    bit hex_suffix = true; /**< Whether to parse hexadecimal suffix (h or H following the number). */
    bit oct = true; /**< Whether to parse octal prefix (0). */
    bit bin = true; /**< Whether to parse binary prefix (0b, 0B). */
    bit underlines = true; /**< Whether to skip underline characters ('_') inside the number. */
    bit exp = true; /**< Whether to allow scientific notation. */
    bit exp_hexbin = false; /**< If true, hexadecimal exponents are read as base 10 and applied as base 2. */
    bit caret_exp = true; /**< Whether to allow the caret character ('^') as a scientific-notation marker. */
    uint radix = 10; /**< Default radix. */
    
    /** Clear all flags to false. */
    void clear ()
    {
        whitespace = sign = hex = hex_suffix = oct = bin = false;
    }
}

const type_flags default_flags;
const type_flags c_flags_int = { true, true, true, false, true, false, false, false, false, false, 10 };
const type_flags c_flags_float = { true, true, false, false, false, false, false, true, false, false, 10 };
const type_flags d_flags_int = { true, true, true, false, true, true, true, false, false, false, 10 };
const type_flags d_flags_float = { true, true, true, false, false, false, true, true, true, false, 10 };
const type_flags csharp_flags_int = { true, true, true, false, false, false, false, false, false, false, 10 };
const type_flags csharp_flags_float = { true, true, false, false, false, false, false, true, false, false, 10 };

private template convert_string (char_type)
{
    struct convert_string
    {
        char_type [] string;
        size_t offset;
        uint radix;
        int digit;
        int sign = 1;
        
        void skip_spaces ()
        {
            while (offset < string.length && isspace (string [offset]))
                offset ++;
        }
        
        bit read_digit ()
        {
            if (offset >= string.length)
                return false;
            char_type ch = string [offset];
            
            while (ch == '_')
            {
                offset ++;
                if (offset >= string.length)
                    return false;
                ch = string [offset];
            }
            
            if (ch >= '0' && ch <= '9')
                digit = ch - '0';
            else if (ch >= 'a' && ch <= 'z')
                digit = ch - 'a' + 10;
            else if (ch >= 'A' && ch <= 'Z')
                digit = ch - 'A' + 10;
            else
                return false;
            if (digit >= radix)
                return false;
            offset ++;
            return true;
        }
        
        bit match (char_type ch)
        {
            if (offset < string.length && string [offset] == ch)
            {
                offset ++;
                return true;
            }
            return false;
        }
        
        bit match (char_type [] match)
        {
            if (offset + match.length <= string.length && string [offset .. offset + match.length] == match)
            {
                offset += match.length;
                return true;
            }
            return false;
        }
        
        void read_prefix (type_flags flags, bit is_float)
        {
            if (flags.whitespace)
                skip_spaces ();
            
            if (flags.sign)
            {
                while (1)
                {
                    if (match ('+'))
                        sign = +sign;
                    else if (match ('-'))
                        sign = -sign;
                    else
                        break;
                }
            }
            
            if (flags.hex && (match ("0x") || match ("0X")))
                radix = 16;
            else if (flags.bin && (match ("0b") || match ("0B")))
                radix = 2;
            else if (flags.oct && match ("0"))
                radix = 8;
            else if (flags.hex_suffix)
            {
                uint original_radix = radix;
                size_t original_offset = offset;
                
                radix = 16;
                while (1)
                {
                    if (is_float && match ("."))
                        continue;
                    else if (match ("h") || match ("H"))
                    {
                        offset = original_offset;
                        break;
                    }
                    else if (!read_digit ())
                    {
                        offset = original_offset;
                        radix = original_radix;
                        break;
                    }
                }
            }
        }
    }
}

template string_to_integer (char_type, int_type)
{
    void string_to_integer (char_type [] string, out int_type result, type_flags flags = default_flags)
    {
        size_t offset;
        
        return string_to_integer (string, result, flags, offset);
    }
    
    void string_to_integer (char_type [] string, out int_type result, type_flags flags, out size_t end_offset)
    in
    {
        assert (flags.radix <= 36);
    }
    body
    {
        convert_string! (char_type) convert;
        int_type whole;
        
        convert.string = string;
        convert.radix = flags.radix;
        convert.read_prefix (flags, false);
 
        bit has_digits = false;
               
        while (convert.read_digit ())
        {
            has_digits = true;
            if (whole * convert.radix + convert.digit < whole)
                throw new class_error_overflow ();
            whole = whole * convert.radix + convert.digit;
        }
        
        if (!has_digits)
            throw new class_error_no_number_found ();
        
        result = whole * convert.sign;
        end_offset = convert.offset;
    }
}

template string_to_float (char_type, float_type)
{
    void string_to_float (char_type [] string, out float_type result, type_flags flags = default_flags)
    {
        uint offset;
        
        return string_to_float (string, result, flags, offset);
    }
    
    void string_to_float (char_type [] string, out float_type result, type_flags flags, out size_t end_offset)
    in
    {
        assert (flags.radix <= 36);
    }
    body
    {
        convert_string! (char_type) convert;
        float_type whole = float_type.max * 0; /* Uses this roundabout way to allow user-defined types. */
        float_type num = whole, den = whole;
        float_type exp = whole;
        int exp_sign = +1;
        uint exp_radix;
        bit has_digits = false;
        
        convert.string = string;
        convert.radix = flags.radix;
        convert.read_prefix (flags, true);
        
        if (!flags.exp_hexbin || convert.radix != 16)
            exp_radix = convert.radix;
        else
            exp_radix = 2;
        
        while (convert.read_digit ())
        {
            has_digits = true;
            if (whole * convert.radix + convert.digit < whole)
                throw new class_error_overflow ();
            whole = whole * convert.radix + convert.digit;
        }
        
        if (convert.match ('.'))
        {
            while (convert.read_digit ())
            {
                has_digits = true;
                if (num * convert.radix + convert.digit < num)
                    throw new class_error_overflow ();
                num = num * convert.radix + convert.digit;
                den = den - 1;
            }
        }
        
        if (has_digits && flags.exp)
        {
            if ((convert.radix < 16 && (convert.match ("e") || convert.match ("E")))
             || (convert.radix > 15 && (convert.match ("p") || convert.match ("P")))
             || (flags.caret_exp && convert.match ("^")))
            {
                uint original_radix = convert.radix;
                
                if (flags.exp_hexbin)
                    convert.radix = 10;
                
                while (1)
                {
                    if (convert.match ('-'))
                        exp_sign = -exp_sign;
                    else if (convert.match ('+'))
                        exp_sign = +exp_sign;
                    else
                        break;
                }
                
                while (convert.read_digit ())
                {
                    if (exp * convert.radix + convert.digit < exp)
                        throw new class_error_overflow ();
                    exp = exp * convert.radix + convert.digit;
                }
                
                convert.radix = original_radix;
            }
        }
        
        if (!has_digits)
            throw new class_error_no_number_found ();

        result = (whole * convert.sign + num * std.math.pow (convert.radix, den)) * std.math.pow (exp_radix, exp * exp_sign);
        end_offset = convert.offset;
    }
}

alias string_to_integer! (char, int).string_to_integer string_to;
alias string_to_integer! (wchar, int).string_to_integer wstring_to;
alias string_to_integer! (dchar, int).string_to_integer dstring_to;

alias string_to_integer! (char, uint).string_to_integer string_to;
alias string_to_integer! (wchar, uint).string_to_integer wstring_to;
alias string_to_integer! (dchar, uint).string_to_integer dstring_to;

alias string_to_integer! (char, long).string_to_integer string_to;
alias string_to_integer! (wchar, long).string_to_integer wstring_to;
alias string_to_integer! (dchar, long).string_to_integer dstring_to;

alias string_to_integer! (char, ulong).string_to_integer string_to;
alias string_to_integer! (wchar, ulong).string_to_integer wstring_to;
alias string_to_integer! (dchar, ulong).string_to_integer dstring_to;

alias string_to_float! (char, float).string_to_float string_to;
alias string_to_float! (wchar, float).string_to_float wstring_to;
alias string_to_float! (dchar, float).string_to_float dstring_to;

alias string_to_float! (char, double).string_to_float string_to;
alias string_to_float! (wchar, double).string_to_float wstring_to;
alias string_to_float! (dchar, double).string_to_float dstring_to;

alias string_to_float! (char, real).string_to_float string_to;
alias string_to_float! (wchar, real).string_to_float wstring_to;
alias string_to_float! (dchar, real).string_to_float dstring_to;

version (test_string_to_type)
{
    void main ()
    {
        type_flags flags;
        uint offset;
        float value;
 
        string_to ("0x10.10p-1", value, d_flags_float);
        writef ("%s %s\n", value, 0x10.10p-1);
    }
}